//! Helper library for msp430-asm
#![feature(decl_macro)]
pub mod split_twice {
    /// Slices a collection into a beginning, middle, and end, based on two unordered indices
    pub trait SplitTwice<'t> {
        type Slice;
        type Idx;
        /// Splits a collection into a beginning, middle, and end slice,
        /// based on two unordered indices
        ///
        /// # Examples
        /// ```rust
        /// # use msp430_asm::split_twice::SplitTwice;
        /// let string = "foo,bar,baz";
        /// let (foo, bar, baz) = string.split_twice(4, 8);
        /// assert_eq!(foo, "foo,");
        /// assert_eq!(bar, "bar,");
        /// assert_eq!(baz, "baz");
        /// ```
        fn split_twice(
            &'t self,
            a: Self::Idx,
            b: Self::Idx,
        ) -> (Self::Slice, Self::Slice, Self::Slice);
    }

    impl<'t, T: 't> SplitTwice<'t> for [T] {
        type Slice = &'t [T];
        type Idx = usize;
        fn split_twice(
            &'t self,
            a: Self::Idx,
            b: Self::Idx,
        ) -> (Self::Slice, Self::Slice, Self::Slice) {
            let (a, b) = if a < b { (a, b) } else { (b, a) };
            let (mid, end) =
                if b < self.len() { self.split_at(b) } else { (self, Default::default()) };
            let (start, mid) =
                if a < mid.len() { mid.split_at(a) } else { (self, Default::default()) };
            (start, mid, end)
        }
    }

    impl<'t> SplitTwice<'t> for str {
        type Slice = &'t str;
        type Idx = usize;
        fn split_twice(
            &'t self,
            a: Self::Idx,
            b: Self::Idx,
        ) -> (Self::Slice, Self::Slice, Self::Slice) {
            let (a, b) = if a < b { (a, b) } else { (b, a) };
            let (mid, end) =
                if b < self.len() { self.split_at(b) } else { (self, Default::default()) };
            let (start, mid) =
                if a < mid.len() { mid.split_at(a) } else { (self, Default::default()) };
            (start, mid, end)
        }
    }
}

pub mod cursor {
    use std::fmt::{Arguments, Display};

    /// Moves to the {line}th previous line
    pub macro previous($line:literal) {
        csi!("{}F", $line)
    }

    /// Injects a Command Sequence Introducer
    pub macro csi($($t:tt)*) {
        format_args!("\x1b[{}", format_args!($($t)*))
    }

    /// Formats the args with a foreground [Color]
    pub macro fg($fg:expr, $($t:tt)*) {
        Colorized::new(Some($fg), None, format_args!($($t)*))
    }

    /// Formats the args with a background [Color]
    pub macro bg($bg:expr, $(t:tt)*) {
        Colorized::new(None, Some($bg), format_args!($($t)*))
    }

    /// Formats the args with both a foreground and background [Color]
    pub macro color($fg:expr, $bg:expr, $($t:tt)*) {
        Colorized::new(Some($fg), Some($bg), format_args!($($t)*))
    }

    #[derive(Clone, Copy, Debug, Default, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub enum Color {
        #[default]
        Black    = 30,
        Red,
        Green,
        Yellow,
        Blue,
        Magenta,
        Cyan,
        Gray,
        DarkGray = 90,
        Pink,
        Lime,
        Sunflower,
        SkyBlue,
        HotPink,
        Turquoise,
        White,
    }

    #[derive(Clone, Copy, Debug)]
    pub struct Colorized<'args> {
        fg: Option<Color>,
        bg: Option<Color>,
        args: Arguments<'args>,
    }

    impl<'t> Colorized<'t> {
        pub fn new(fg: Option<Color>, bg: Option<Color>, args: Arguments<'t>) -> Self {
            Self { fg, bg, args }
        }
    }
    impl<'t> Display for Colorized<'t> {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            let &Self { fg, bg, args } = self;
            if let Some(fg) = fg {
                write!(f, "{}", csi!("{}m", fg as u8))?;
            }
            if let Some(bg) = bg {
                write!(f, "{}", csi!("{}m", bg as u8 + 10))?;
            }
            write!(f, "{args}{}", csi!("0m"))
        }
    }
}
